
type var = string

type typ =
    Bool
  | Int
  | Arrow of typ * typ (* Arrow(t1,t2) means function from t1 to t2 *)
  | Ignore (* place-holder used for functions built during evaluation *)

type context = (var * typ) list

type exp =
    True
  | False
  | If of exp * exp * exp
  | Const of int
  | Plus of exp * exp
  | GreaterThan of exp * exp
  | Var of var
  | Fun of var * typ * exp
  | Apply of exp * exp
  | Rec of var * typ * var * exp (*typ is the type of the recursive function*)
  | Closure of exp * env
(*and env = var -> exp
 other implementation:*)
 and env = (var * (exp ref)) list

(* for debugging purposes -- not tested -- buyer beware *)

let rec typ2string t =
  match t with
    Bool -> "bool"
  | Int -> "int"
  | Arrow(t1,t2) -> "(" ^ (typ2string t1) ^ "->" ^ (typ2string t2) ^ ")"
  | Ignore -> "_"

let rec exp2string e =
  match e with
    True  -> "true"
  | False -> "false"
  | If(e1,e2,e3) -> "(if "   ^ (exp2string e1) ^ 
                    " then " ^ (exp2string e2) ^ 
                    " else " ^ (exp2string e3) ^ ")"
  | Const i -> string_of_int i
  | Plus(e1,e2) -> "(" ^ (exp2string e1) ^ "+" ^ (exp2string e2) ^ ")"
  | GreaterThan(e1,e2) -> "(" ^ (exp2string e1) ^ ">" ^ (exp2string e2) ^ ")"
  | Var(v) -> v
  | Fun(x,t,e) ->
      "(fun " ^ x ^ " : " ^ (typ2string t) ^ " . " ^ (exp2string e) ^ ")"
  | Closure(e,env) ->
      "<" ^ (exp2string e) ^ " , " ^ "_" ^ ">" (* see comment below *)
  | Apply(e1,e2) ->
      (exp2string e1) ^ " " ^ (exp2string e2)
  | Rec(f,t,x,e) ->
      "(rec " ^ f ^ " : " ^ (typ2string t) ^ " " ^ x ^ " . " 
      ^ (exp2string e) ^ ")"

(* We can print environments when using the list representation, but an
   environment containing the closure for a recursive function would
   print an infinite string.  So exp2string does not print environments.

   This can be avoided using pointer equality to detect a cycle, but
   the code below (commented out) does not do this. 
   
   Moreover, when using the function representation, it is impossible to
   print environments.
*)
 
(*and env2string env = 
  match env with
    [] -> "."
  | (x,r)::tl -> (env2string tl) ^ "," ^ x ^ " --> " ^ (exp2string !r)
*)
